#include "StdAfx.h"
#include "resource.h"
#include "LeoHelpers.h"
#include "ModifySetPointConfig.h"

void ModifySetPointConfig::Main(HINSTANCE hInstance, HWND hwndDlg, bool bUndo)
{
	wchar_t szTempPath[MAX_PATH + 1] = {0};

	if (S_OK != SHGetFolderPath(hwndDlg, CSIDL_APPDATA, NULL, SHGFP_TYPE_CURRENT, szTempPath))
	{
		MessageBox(hwndDlg, L"Couldn't find AppData directory.", SPAON_MB_TITLE, MB_OK | MB_ICONERROR);
		return;
	}

	std::wstring strAppData = szTempPath;
	strAppData += L"\\Logitech\\SetPoint";

	if (S_OK != SHGetFolderPath(hwndDlg, CSIDL_COMMON_APPDATA, NULL, SHGFP_TYPE_CURRENT, szTempPath))
	{
		MessageBox(hwndDlg, L"Couldn't find ProgramData directory.", SPAON_MB_TITLE, MB_OK | MB_ICONERROR);
		return;
	}

	std::wstring strProgramData = szTempPath;
	strProgramData += L"\\Logitech\\SetPoint\\Devices\\PointingDevice";

	HWND hwndStatic = GetDlgItem(hwndDlg, IDC_STATIC_DESCRIPTION);
	HWND hwndEdit   = GetDlgItem(hwndDlg, IDC_EDIT_OUTPUT);

	if (hwndStatic == NULL
	||	hwndEdit == NULL)
	{
		MessageBox(hwndDlg, L"Error setting up GUI.", SPAON_MB_TITLE, MB_OK | MB_ICONERROR);
		return;
	}

	if (!bUndo
	&&	IDOK != MessageBox(hwndDlg, L"Warning: Using SetPoint to modify your configuration after using this program may make the changes permanent, so that the Undo option has no effect. You may have to reinstall SetPoint to restore the defaults.", SPAON_MB_TITLE, MB_OKCANCEL | MB_ICONWARNING))
	{
		AppendEditLine(hwndEdit, L"Cancelled.");
		AppendEditLine(hwndEdit, L"");
		return;
	}

	if (bUndo)
	{
		AppendEditLine(hwndEdit, L"Restoring SetPoint mouse application overrides:");
	}
	else
	{
		AppendEditLine(hwndEdit, L"Removing SetPoint mouse application overrides:");
	}

	int iDoneFiles = 0;

	std::wstring strSub;
	std::wstring strPDSubs = strProgramData;
	strPDSubs += L"\\*";

	WIN32_FIND_DATA findData = {0};

	HANDLE hFind = FindFirstFile(strPDSubs.c_str(), &findData);

	if (hFind ==  INVALID_HANDLE_VALUE)
	{
		MessageBoxFileError(hwndDlg, hwndEdit, strPDSubs.c_str(), L"Failed to search", NULL);
		AppendSummaryLine(hwndEdit, iDoneFiles, false);
		return;
	}
	else
	{
		do
		{
			if ((findData.dwFileAttributes & FILE_ATTRIBUTE_DIRECTORY)
			&&	0 != wcscmp(L".",  findData.cFileName)
			&&	0 != wcscmp(L"..", findData.cFileName))
			{
				strSub = strProgramData;
				strSub += L"\\";
				strSub += findData.cFileName;

				if (!ProcessDirectory(hwndDlg, hwndEdit, strSub.c_str(), bUndo, &iDoneFiles))
				{
					AppendSummaryLine(hwndEdit, iDoneFiles, false);
					FindClose(hFind);
					return;
				}
			}
		}
		while(FindNextFile(hFind, &findData));

		if (GetLastError() != ERROR_NO_MORE_FILES)
		{
			MessageBoxFileError(hwndDlg, hwndEdit, strPDSubs.c_str(), L"Failed to search", NULL);
			AppendSummaryLine(hwndEdit, iDoneFiles, false);
			FindClose(hFind);
			return;
		}

		FindClose(hFind);
	}

	// Do the private config last so that it isn't modified if the global config needs editing but
	// the user running the tool doesn't have access to do so.

	std::wstring strUserFilePath = strAppData;
	strUserFilePath += L"\\user.xml";

	if (!ProcessFile(hwndDlg, hwndEdit, strUserFilePath, bUndo, &iDoneFiles))
	{
		AppendSummaryLine(hwndEdit, iDoneFiles, false);
		return;
	}

	AppendSummaryLine(hwndEdit, iDoneFiles, true);

	if (iDoneFiles != 0)
	{
		const wchar_t *szMsg = L"You should logout and back in again to restart SetPoint with the new configuration.";

		AppendEditLine(hwndEdit, szMsg);

		MessageBox(hwndDlg, szMsg, SPAON_MB_TITLE, MB_OK | MB_ICONINFORMATION);
	}
}

void ModifySetPointConfig::AppendEditLine(HWND hwndEdit, const wchar_t *szText)
{
	int ndx;

	ndx = GetWindowTextLength(hwndEdit);
	SendMessage(hwndEdit, EM_SETSEL, ndx, ndx);
	SendMessage(hwndEdit, EM_REPLACESEL, FALSE, reinterpret_cast< LPARAM >( szText ));

	ndx = GetWindowTextLength(hwndEdit);
	SendMessage(hwndEdit, EM_SETSEL, ndx, ndx);
	SendMessage(hwndEdit, EM_REPLACESEL, FALSE, reinterpret_cast< LPARAM >( L"\r\n" ));

	ndx = GetWindowTextLength(hwndEdit);
	SendMessage(hwndEdit, EM_SETSEL, ndx, ndx);
	SendMessage(hwndEdit, EM_SCROLLCARET, 0, 0);
}

void ModifySetPointConfig::AppendSummaryLine(HWND hwndEdit, int iDoneFiles, bool bSuccess)
{
	if (iDoneFiles == 0)
	{
		if (bSuccess)
		{
			AppendEditLine(hwndEdit, L"DONE. No modifications were needed.");
		}
		else
		{
			AppendEditLine(hwndEdit, L"FAILED. No modifications succeeded.");
		}
	}
	else
	{
		std::wstring strMsg;

		if (bSuccess)
		{
			LeoHelpers::StringFormatToString(&strMsg, L"DONE. %d files were modified.", iDoneFiles);
		}
		else
		{
			LeoHelpers::StringFormatToString(&strMsg, L"FAILED. However, %d files were modified successfully.", iDoneFiles);
		}

		AppendEditLine(hwndEdit, strMsg.c_str());
	}

	AppendEditLine(hwndEdit, L"");
}

void ModifySetPointConfig::MessageBoxFileError(HWND hwndDlg, HWND hwndEdit, const wchar_t *szFilePath, const wchar_t *szErrMsg, const wchar_t *szAdditionalMsg)
{
	std::wstring strMsg = szErrMsg;
	strMsg += L" \"";
	strMsg += szFilePath;
	strMsg += L"\"";
	AppendEditLine(hwndEdit, strMsg.c_str());

	if (szAdditionalMsg)
	{
		AppendEditLine(hwndEdit, szAdditionalMsg);

		strMsg += L"\n\n";
		strMsg += szAdditionalMsg;
	}

	MessageBox(hwndDlg, strMsg.c_str(), SPAON_MB_TITLE, MB_OK | MB_ICONERROR);
}

bool ModifySetPointConfig::ProcessDirectory(HWND hwndDlg, HWND hwndEdit, const std::wstring &strFilePath, bool bUndo, int *piDoneFiles)
{
	std::wstring strFilePattern = strFilePath;
	strFilePattern += L"\\*.xml";

	WIN32_FIND_DATA findData = {0};

	HANDLE hFind = FindFirstFile(strFilePattern.c_str(), &findData);

	if (hFind == INVALID_HANDLE_VALUE)
	{
		if (GetLastError() != ERROR_NO_MORE_FILES)
		{
			return true; // Some of the folders don't contain any .xml files.
		}

		MessageBoxFileError(hwndDlg, hwndEdit, strFilePattern.c_str(), L"Failed to search", NULL);
		return false;
	}
	else
	{
		std::wstring strSub;

		do
		{
			if (!(findData.dwFileAttributes & FILE_ATTRIBUTE_DIRECTORY))
			{
				strSub = strFilePath;
				strSub += L"\\";
				strSub += findData.cFileName;

				if (!ProcessFile(hwndDlg, hwndEdit, strSub.c_str(), bUndo, piDoneFiles))
				{
					FindClose(hFind);
					return false;
				}
			}
		}
		while(FindNextFile(hFind, &findData));

		if (GetLastError() != ERROR_NO_MORE_FILES)
		{
			MessageBoxFileError(hwndDlg, hwndEdit, strFilePattern.c_str(), L"Failed to search", NULL);
			FindClose(hFind);
			return false;
		}

		FindClose(hFind);
	}

	return true;
}

bool ModifySetPointConfig::ProcessFile(HWND hwndDlg, HWND hwndEdit, const std::wstring &strFilePath, bool bUndo, int *piDoneFiles)
{
	bool bModified = false;

	if (ProcessFileInner(hwndDlg, hwndEdit, strFilePath.c_str(), bUndo, &bModified))
	{
		if (bModified)
		{
			std::wstring strMsg = L"MODIFIED: ";
			strMsg += strFilePath;
			AppendEditLine(hwndEdit, strMsg.c_str());

			++(*piDoneFiles);
		}

		return true;
	}

	std::wstring strMsg = L"FAILED: ";
	strMsg += strFilePath;
	AppendEditLine(hwndEdit, strMsg.c_str());

	AppendEditLine(hwndEdit, L"Aborted.");
	return false;
}

bool ModifySetPointConfig::ProcessFileInner(HWND hwndDlg, HWND hwndEdit, const wchar_t *szFilePath, bool bUndo, bool *pbModified)
{
	bool bResult = false;
	*pbModified = false;

	bool bIsReadOnly = false;

	HANDLE hFile = CreateFile(szFilePath, FILE_READ_ATTRIBUTES|FILE_READ_DATA|FILE_APPEND_DATA|FILE_WRITE_DATA, 0, NULL,
								OPEN_EXISTING, FILE_ATTRIBUTE_NORMAL|FILE_FLAG_SEQUENTIAL_SCAN, NULL);

	// If we failed to open the file due to lack of access, try again only requesting read access.
	// This way if a (non-admin) user runs the tool we'll only error for them on files that we actually need
	// to modify by can't, so that if the common settings files have already modified by an admin a standard user
	// can then use the tool to update just his own private config file, without error. However, if the common
	// files require modification & the user doesn't have access then we will consider that a failure.

	if (hFile == INVALID_HANDLE_VALUE && GetLastError() == ERROR_ACCESS_DENIED)
	{
		bIsReadOnly = true;

		hFile = CreateFile(szFilePath, FILE_READ_ATTRIBUTES|FILE_READ_DATA, 0, NULL,
							OPEN_EXISTING, FILE_ATTRIBUTE_NORMAL|FILE_FLAG_SEQUENTIAL_SCAN, NULL);
	}

	LARGE_INTEGER liFileSize = {0};

	// liFileSize.HighPart check is because it'd be ridiculous for one of these XML files to be over 4gig
	// and I don't want to complicate the code by having to call ReadFile in a loop.

	if (hFile == INVALID_HANDLE_VALUE || !GetFileSizeEx(hFile, &liFileSize) || liFileSize.HighPart != 0)
	{
		MessageBoxFileError(hwndDlg, hwndEdit, szFilePath, L"Failed to open", NULL);
	}
	else
	{
		char *fileBuf = new(std::nothrow) char[liFileSize.LowPart];

		if (fileBuf == NULL)
		{
			MessageBoxFileError(hwndDlg, hwndEdit, szFilePath, L"Failed to allocate memory for", NULL);
		}
		else
		{
			DWORD dwBytesDone = 0;

			if (!ReadFile(hFile, fileBuf, liFileSize.LowPart, &dwBytesDone, NULL) || dwBytesDone != liFileSize.LowPart)
			{
				MessageBoxFileError(hwndDlg, hwndEdit, szFilePath, L"Failed to read", NULL);
			}
			else
			{
				bool bModified = false;

				if (FindRep(hwndDlg, fileBuf, liFileSize.LowPart, bUndo ? "<NUKEDOVRIDE" : "<APPOVERRIDE", bUndo ? "<AppOverride" : "<NUKEDOVRIDE"))
				{
					bModified = true;
				}

				if (FindRep(hwndDlg, fileBuf, liFileSize.LowPart, bUndo ? "</NUKEDOVRIDE" : "</APPOVERRIDE", bUndo ? "</AppOverride" : "</NUKEDOVRIDE"))
				{
					bModified = true;
				}

				if (FindRep(hwndDlg, fileBuf, liFileSize.LowPart, bUndo ? "<NUK" : "<APP", bUndo ? "<App" : "<NUK"))
				{
					bModified = true;
				}

				if (FindRep(hwndDlg, fileBuf, liFileSize.LowPart, bUndo ? "</NUK" : "</APP", bUndo ? "</App" : "</NUK"))
				{
					bModified = true;
				}

				if (!bModified)
				{
					bResult = true;
				}
				else if (bIsReadOnly)
				{
					MessageBoxFileError(hwndDlg, hwndEdit, szFilePath, L"Cannot write to", L"The first user to run the tool must be an administrator so that the global configuration files may be modified.");
				}
				else
				{
					if (SetFilePointer(hFile, 0, NULL, FILE_BEGIN) == INVALID_SET_FILE_POINTER
					||	!WriteFile(hFile, fileBuf, liFileSize.LowPart, &dwBytesDone, NULL)
					||	dwBytesDone != liFileSize.LowPart)
					{
						MessageBoxFileError(hwndDlg, hwndEdit, szFilePath, L"Failed to update", NULL);
					}
					else
					{
						*pbModified = true;
						bResult = true;
					}
				}
			}

			delete[] fileBuf;
		}
	}

	if (hFile != INVALID_HANDLE_VALUE)
	{
		CloseHandle(hFile);
	}

	return bResult;
}

bool ModifySetPointConfig::FindRep(HWND hwndDlg, char *fileBuf, DWORD dwFileSize, const char *szSearchUpper, const char *szReplace)
{
	bool bModified = false;

	char *pBuf = fileBuf;
	DWORD dwRemaining = dwFileSize;
	const size_t lenSearch = strlen(szSearchUpper);

	if (strlen(szReplace) != lenSearch)
	{
		MessageBox(hwndDlg, L"Internal Error", SPAON_MB_TITLE, MB_OK | MB_ICONERROR);
		return false;
	}

	while (dwRemaining >= (lenSearch+1))
	{
		if (0 != LeoHelpers::StrCmpUpperLenLeftOnly(pBuf, szSearchUpper, lenSearch)
		||	(pBuf[lenSearch] != ' ' && pBuf[lenSearch] != '>'))
		{
			++pBuf;
			--dwRemaining;
			continue;
		}

		bModified = true;
		memcpy(pBuf, szReplace, lenSearch);
		pBuf += lenSearch;
		dwRemaining -= lenSearch;
	}

	return bModified;
}